package net.varek.rs2d;

import java.io.IOException;
import java.util.List;
import org.apache.mina.common.IoSession;
import net.varek.rs2d.model.Player;
import net.varek.rs2d.model.Tile;
import net.varek.rs2d.model.World;
import net.varek.rs2d.net.ActionSender;
import net.varek.rs2d.net.Packet;
import net.varek.rs2d.net.PacketQueue;
import net.varek.rs2d.packetbuilder.StaticPacketBuilder;
import net.varek.rs2d.packethandler.PacketHandler;
import net.varek.rs2d.util.Logger;

/**
 * The central motor of the game. This class is responsible for the
 * primary operation of the entire game.
 */
public final class GameEngine implements Runnable {
	/**
	 * The packet queue to be processed
	 */
	private PacketQueue packetQueue = new PacketQueue();
	/**
	 * Whether the engine's thread is running
	 */
	private boolean running = true;
	/**
	 * The mapping of packet IDs to their handler
	 */
	private PacketHandler[] packetHandlers = new PacketHandler[256];
	/**
	 * Responsible for updating all connected clients
	 */
	private ClientUpdater clientUpdater = new ClientUpdater();
	/**
	 * Handles delayed events rather than events to be ran every iteration
	 */
	private DelayedEventHandler eventHandler = new DelayedEventHandler();
	/**
	 * NPC Updater. See class for documentation
	 */

	/**
	 * Constructs a new game engine with an empty packet queue.
	 */
	public GameEngine() throws IOException {
		
	}

	/**
	 * The thread execution process.
	 */
	public void run() {
		long lastSentEventUpdate = 0, lastSentMajorUpdate = 0, lastSentMinorUpdate = 0, 
			lastTileCleanup = System.currentTimeMillis(), lastPerfInfo = 0;
		while (running) {
			try {
				Thread.sleep(30);
				processIncomingPackets();
				processPlayersAwaitingLogin();
				long curTime = System.currentTimeMillis();
				if(curTime - lastSentEventUpdate >= 100) {
					lastSentEventUpdate = curTime;
					eventHandler.doEvents();
				}
				if(curTime - lastSentMajorUpdate >= 500) {
					lastSentMajorUpdate = curTime;
					clientUpdater.majorUpdate();
				}
				if(curTime - lastSentMinorUpdate >= 200) {
					lastSentMinorUpdate = curTime;
					clientUpdater.minorUpdate();
				}
				if(curTime - lastTileCleanup >= 60 * 8 * 1000) {
					lastTileCleanup = curTime;
					int cleaned = cleanUpTiles(); //ideally this should be done asynchronously.
				}
			} catch(Exception e) {
				lastSentEventUpdate = lastSentMinorUpdate = lastSentMajorUpdate = 0L;
				Logger.err("Error processing updates,");
				Logger.err(e);
			}
		}
	}
	
	public int cleanUpTiles() {
		Tile[][] tiles = World.getWorld().tiles;
		int cleaned = 0;
		for(int x=0; x<tiles.length; x++) {
			for(int y=0; y<tiles[0].length; y++) {
				Tile t = tiles[x][y];
				if(t == null) continue;
				if(t.isEmpty()) {
					t = null;
					cleaned++;
				}
			}
		}
		return cleaned;
	}

	/**
	 * Returns the current packet queue.
	 *
	 * @return A <code>PacketQueue</code>
	 */
	public PacketQueue getPacketQueue() {
		return packetQueue;
	}

	/**
	 * Loads the packet handling classes from the persistence
	 * manager.
	 */
	public void loadPacketHandlers() throws IOException {
		
	}

	/**
	 * Processes incoming packets.
	 */
	public void processIncomingPackets() {
		for(Packet p : packetQueue.getPackets()) {
			IoSession session = p.getSession();
			Player player = (Player)(session.getAttachment());
			Logger.log("Unhandled: "+p);
		}
	}
	
	/**
	 * Process all players waiting to be logged in
	 */
	public void processPlayersAwaitingLogin() {
		if(World.getWorld().getPlayersAwaitingLogin().isEmpty())
			return;
		Player[] players = null;
		List<Player> worldList = World.getWorld().getPlayersAwaitingLogin();
		synchronized(worldList) {
			players = worldList.toArray(new Player[worldList.size()]);
			worldList.clear();
		}
		for(Player player : players) {
			byte loginCode = player.getResponseCode();
			final IoSession session = player.getIoSession();
			try {
				int loginResponse = player.getResponseCode();
				//Logger.log("resp: "+loginResponse);
				switch(loginResponse) {
				case 2:
					ActionSender sender = player.getActionSender();
					sender.sendLoginResponse(2, 0, 0);
					sender.sendLoginPacket();
					sender.sendPrivacySettings();
					for(int i=0; i<25; i++)
						sender.sendStat(i);
					sender.sendUnknown107();
					
					int[][] sideBarInterfaces = {
						{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 0},
						{3917, 638, 3213, 1644, 5608, 1151, 1, 5065, 5715, 2449, 4445, 147, 6299, 2423}
					};
					for(int i=0; i<sideBarInterfaces[0].length; i++)
						sender.sendSideBarInterface(sideBarInterfaces[0][i], sideBarInterfaces[1][i]);
					
					sender.sendPlayerCommand(3, true, "Trade with");
					//sender.sendLoginWindow(201, false, 3, (127 << 24)+1, (int)(Math.random() * 10));
					
					player.getIoSession().write(new StaticPacketBuilder().setId(73).addShortA(406).addShort(428).toPacket());
					
					World.getWorld().registerPlayer(player);
				}
			} catch(Exception e) {
				Logger.err("Error loading "+player);
				Logger.err(e);
			}
		}
	}
}
